/*
 * Copyright 1999-2016 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the Apache License 2.0 (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

#ifndef OSSL_APPS_TIMEOUTS_H
# define OSSL_APPS_TIMEOUTS_H

/* numbers in us */
# define DGRAM_RCV_TIMEOUT         250000
# define DGRAM_SND_TIMEOUT         250000

#endif                          /* ! OSSL_APPS_TIMEOUTS_H */
